@@ -1,6 +1,7 @@
 /* Declarations for connect.
    Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-   2005, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
+   2005, 2006, 2007, 2008, 2009, 2010, 2011 Free Software Foundation,
+   Inc.
 
 This file is part of GNU Wget.
 
@@ -50,6 +51,7 @@ enum {
   ENDPOINT_PEER
 };
 bool socket_ip_address (int, ip_address *, int);
+int  socket_family (int sock, int endpoint);
 
 bool retryable_socket_connect_error (int);
 
@@ -58,7 +60,11 @@ enum {
   WAIT_FOR_READ = 1,
   WAIT_FOR_WRITE = 2
 };
+#ifdef ENABLE_THREADS
+int select_fds (int *, int *, int, double, int);
+#else
 int select_fd (int, double, int);
+#endif
 bool test_socket_open (int);
 
 struct transport_implementation {
