@@ -1,6 +1,6 @@
 /* Handling of recursive HTTP retrieving.
    Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-   2005, 2006, 2007, 2008, 2009, 2010, 2011 Free Software Foundation,
+   2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012 Free Software Foundation,
    Inc.
 
 This file is part of GNU Wget.
@@ -37,7 +37,12 @@ as that of the covered work.  */
 #include <unistd.h>
 #include <errno.h>
 #include <assert.h>
+#ifdef ENABLE_THREADS
+#include <pthread.h>
+#include <semaphore.h>
 
+#include "multi.h"
+#endif
 #include "url.h"
 #include "recur.h"
 #include "utils.h"
@@ -165,6 +170,20 @@ static bool download_child_p (const struct urlpos *, struct url *, int,
 static bool descend_redirect_p (const char *, struct url *, int,
                                 struct url *, struct hash_table *, struct iri *);
 
+#ifdef ENABLE_THREADS
+static void *
+start_retrieve_url (void *arg)
+{
+  struct s_thread_ctx *ctx = (struct s_thread_ctx *) arg;
+
+  ctx->status = retrieve_url (ctx->url_parsed, ctx->url,
+                              &ctx->file, &ctx->redirected,
+                              ctx->referer, &ctx->dt,
+                              false, ctx->i, true, NULL);
+  ctx->terminated = 1;
+  sem_post (ctx->retr_sem);
+}
+#endif
 
 /* Retrieve a part of the web beginning with START_URL.  This used to
    be called "recursive retrieval", because the old function was
@@ -191,6 +210,17 @@ uerr_t
 retrieve_tree (struct url *start_url_parsed, struct iri *pi)
 {
   uerr_t status = RETROK;
+#ifdef ENABLE_THREADS
+  const int N_THREADS = opt.jobs > 0 ? opt.jobs : 1;
+  sem_t retr_sem;
+  int free_threads = N_THREADS;
+  char *next_url = NULL, *next_referer;
+  int next_depth;
+  bool next_html_allowed, next_css_allowed;
+  struct iri *next_i = NULL;
+
+  struct s_thread_ctx *thread_ctx;
+#endif
 
   /* The queue of URLs we need to load. */
   struct url_queue *queue;
@@ -213,6 +243,12 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
     set_uri_encoding (i, opt.locale, true);
 #undef COPYSTR
 
+#ifdef ENABLE_THREADS
+  thread_ctx = calloc (N_THREADS, sizeof *thread_ctx);
+  sem_init (&retr_sem, 0, 0);
+  /* FIXME: CHECK FOR ERRORS.  */
+#endif
+
   queue = url_queue_new ();
   blacklist = make_string_hash_table (0);
 
@@ -225,11 +261,16 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
   while (1)
     {
       bool descend = false;
-      char *url, *referer, *file = NULL;
-      int depth;
-      bool html_allowed, css_allowed;
+      char *file = NULL;
       bool is_css = false;
       bool dash_p_leaf_HTML = false;
+      char *url = NULL, *referer;
+      int depth;
+      bool html_allowed, css_allowed;
+      bool dequed = false;
+#ifdef ENABLE_THREADS
+      int index = 0;
+#endif
 
       if (opt.quota && total_downloaded_bytes > opt.quota)
         break;
@@ -238,10 +279,27 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
 
       /* Get the next URL from the queue... */
 
+#ifndef ENABLE_THREADS
       if (!url_dequeue (queue, (struct iri **) &i,
                         (const char **)&url, (const char **)&referer,
                         &depth, &html_allowed, &css_allowed))
         break;
+#else
+      if (next_url == NULL)
+        {
+          if (url_dequeue (queue, (struct iri **) &next_i,
+                           (const char **)&next_url, (const char **)&next_referer,
+                           &next_depth, &next_html_allowed, &next_css_allowed))
+            dequed = true;
+        }
+
+      i = next_i;
+      url = next_url;
+      referer = next_referer;
+      depth = next_depth;
+      html_allowed = next_html_allowed;
+      css_allowed = next_css_allowed;
+#endif
 
       /* ...and download it.  Note that this download is in most cases
          unconditional, as download_child_p already makes sure a file
@@ -251,40 +309,131 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
          and again under URL2, but at a different (possibly smaller)
          depth, we want the URL's children to be taken into account
          the second time.  */
-      if (dl_url_file_map && hash_table_contains (dl_url_file_map, url))
+      if (dequed && url && dl_url_file_map && hash_table_contains (dl_url_file_map, url))
         {
+	  bool is_css_bool;
+
           file = xstrdup (hash_table_get (dl_url_file_map, url));
 
           DEBUGP (("Already downloaded \"%s\", reusing it from \"%s\".\n",
                    url, file));
 
-          /* this sucks, needs to be combined! */
-          if (html_allowed
-              && downloaded_html_set
-              && string_set_contains (downloaded_html_set, file))
-            {
-              descend = true;
-              is_css = false;
-            }
-          if (css_allowed
-              && downloaded_css_set
-              && string_set_contains (downloaded_css_set, file))
-            {
-              descend = true;
-              is_css = true;
-            }
+	  if ((is_css_bool = (css_allowed
+			      && downloaded_css_set
+			      && string_set_contains (downloaded_css_set, file)))
+	      || (html_allowed
+		  && downloaded_html_set
+		  && string_set_contains (downloaded_html_set, file)))
+	    {
+	      descend = true;
+	      is_css = is_css_bool;
+	    }
         }
       else
         {
+#ifndef ENABLE_THREADS
           int dt = 0, url_err;
           char *redirected = NULL;
           struct url *url_parsed = url_parse (url, &url_err, i, true);
 
           status = retrieve_url (url_parsed, url, &file, &redirected, referer,
-                                 &dt, false, i, true);
+                                 &dt, false, i, true, NULL);
+#else
+          int j;
+retry:
+          if (! url)
+            {
+              int used = 0;
+              for (j = 0; j < N_THREADS; j++)
+                {
+                  if (thread_ctx[j].used)
+                    {
+                      used = 1;
+                      break;
+                    }
+                }
+
+              if (! used)
+                break;
+            }
+
+          if (url && free_threads)
+            {
+              int err;
+              for (j = 0; j < N_THREADS; j++)
+                if (! thread_ctx[j].used)
+                  {
+                    index = j;
+                    free_threads--;
+                    thread_ctx[j].used = 1;
+                    thread_ctx[j].terminated = 0;
+                    break;
+                  }
+              assert (j < N_THREADS);
+
+              thread_ctx[index].file = file;
+              thread_ctx[index].referer = referer;
+              thread_ctx[index].dt = 0;
+              thread_ctx[index].i = i;
+              thread_ctx[index].redirected = NULL;
+              thread_ctx[index].range = NULL;
+              thread_ctx[index].url = url;
+              thread_ctx[index].retr_sem = &retr_sem;
+              thread_ctx[index].url_parsed = url_parse (thread_ctx[index].url,
+                                                        &thread_ctx[index].url_err,
+                                                        i, true);
+
+              err = pthread_create (&thread_ctx[index].thread, NULL,
+                                    start_retrieve_url, &thread_ctx[index]);
+              if (err == 0)
+                next_url = NULL;
+              else
+                {
+                  logprintf (LOG_NOTQUIET, "pthread_create: %s\n", strerror (errno));
+                  url_free (thread_ctx[index].url_parsed);
+                  thread_ctx[index].used = 0;
+                  free_threads++;
+                  return THREADS_ERR;
+                }
+              continue;
+            }
+
+          index = -1;
+          for (j = 0; j < N_THREADS; j++)
+            if (thread_ctx[j].used && thread_ctx[j].terminated)
+              {
+                index = j;
+                thread_ctx[j].used = 0;
+                pthread_join (thread_ctx[j].thread, NULL);
+                free_threads++;
+                break;
+              }
+
+          if (index < 0)
+            {
+              int ret;
+              do
+                ret = sem_wait (&retr_sem);
+              while (ret < 0 && errno == EINTR);
+              if (ret < 0)
+                return SEM_ERR;
+
+              goto retry;
+            }
+
+          file = thread_ctx[index].file;
+          referer = thread_ctx[index].referer;
+          i = thread_ctx[index].i;
+          url = thread_ctx[index].url;
+#endif
 
           if (html_allowed && file && status == RETROK
-              && (dt & RETROKF) && (dt & TEXTHTML))
+#ifndef ENABLE_THREADS
+              && (dt & RETROKF) && (dt & TEXTHTML)
+#else
+              && (thread_ctx[index].dt & RETROKF) && (thread_ctx[index].dt & TEXTHTML)
+#endif
+             )
             {
               descend = true;
               is_css = false;
@@ -294,21 +443,36 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
              lots of web servers serve css with an incorrect content type
           */
           if (file && status == RETROK
+#ifndef ENABLE_THREADS
               && (dt & RETROKF) &&
-              ((dt & TEXTCSS) || css_allowed))
+              ((dt & TEXTCSS) || css_allowed)
+#else
+              && (thread_ctx[index].dt & RETROKF) &&
+              ((thread_ctx[index].dt & TEXTCSS) || css_allowed)
+#endif
+             )
             {
               descend = true;
               is_css = true;
             }
 
+#ifndef ENABLE_THREADS
           if (redirected)
+#else
+          if (thread_ctx[index].redirected)
+#endif
             {
               /* We have been redirected, possibly to another host, or
                  different path, or wherever.  Check whether we really
                  want to follow it.  */
               if (descend)
                 {
+#ifndef ENABLE_THREADS
                   if (!descend_redirect_p (redirected, url_parsed, depth,
+#else
+                  if (!descend_redirect_p (thread_ctx[index].redirected,
+                                           thread_ctx[index].url_parsed, depth,
+#endif
                                            start_url_parsed, blacklist, i))
                     descend = false;
                   else
@@ -317,15 +481,29 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
                     string_set_add (blacklist, url);
                 }
 
+#ifndef ENABLE_THREADS
               xfree (url);
               url = redirected;
+#else
+              xfree (thread_ctx[index].url);
+              url = thread_ctx[index].redirected;
+#endif
             }
           else
             {
+#ifndef ENABLE_THREADS
               xfree (url);
               url = xstrdup (url_parsed->url);
+#else
+              xfree (thread_ctx[index].url);
+              url = xstrdup (thread_ctx[index].url_parsed->url);
+#endif
             }
+#ifndef ENABLE_THREADS
           url_free(url_parsed);
+#else
+          url_free(thread_ctx[index].url_parsed);
+#endif
         }
 
       if (opt.spider)
@@ -441,11 +619,12 @@ retrieve_tree (struct url *start_url_parsed, struct iri *pi)
           logputs (LOG_VERBOSE, "\n");
           register_delete_file (file);
         }
-
+#ifndef ENABLE_THREADS
       xfree (url);
       xfree_null (referer);
       xfree_null (file);
       iri_free (i);
+#endif
     }
 
   /* If anything is left of the queue due to a premature exit, free it
@@ -586,6 +765,11 @@ download_child_p (const struct urlpos *upos, struct url *parent, int depth,
           goto out;
         }
     }
+  if (!accept_url (url))
+    {
+      DEBUGP (("%s is excluded/not-included through regex.\n", url));
+      goto out;
+    }
 
   /* 6. Check for acceptance/rejection rules.  We ignore these rules
      for directories (no file name to match) and for non-leaf HTMLs,
@@ -636,7 +820,7 @@ download_child_p (const struct urlpos *upos, struct url *parent, int depth,
                  files after downloading or we're just running a spider. */
               if (opt.delete_after || opt.spider)
                 {
-                  logprintf (LOG_VERBOSE, "Removing %s.\n", rfile);
+                  logprintf (LOG_VERBOSE, _("Removing %s.\n"), rfile);
                   if (unlink (rfile))
                       logprintf (LOG_NOTQUIET, "unlink: %s\n",
                                  strerror (errno));
