@@ -314,6 +314,10 @@ typedef double SUM_SIZE_INT;
    0-pad the address.)  */
 #define PTR_FORMAT(p) (int) (2 * sizeof (void *)), (unsigned long) (p)
 
+/* Find the maximum buffer length needed to print an integer of type `x'
+   in base 10. 24082 / 10000 = 8*log_{10}(2).  */
+#define MAX_INT_TO_STRING_LEN(x) ((sizeof(x) * 24082 / 10000) + 2)
+
 extern const char *exec_name;
 
 /* Document type ("dt") flags */
@@ -353,11 +357,21 @@ typedef enum
   PROXERR,
   /* 50  */
   AUTHFAILED, QUOTEXC, WRITEFAILED, SSLINITFAILED, VERIFCERTERR,
-  UNLINKERR, NEWLOCATION_KEEP_POST, CLOSEFAILED,
-
-  WARC_ERR, WARC_TMP_FOPENERR, WARC_TMP_FWRITEERR
+  UNLINKERR, NEWLOCATION_KEEP_POST, CLOSEFAILED, WARC_ERR,
+  WARC_TMP_FOPENERR,
+  /* 60  */
+  WARC_TMP_FWRITEERR, THREADS_ERR, SEM_ERR
 } uerr_t;
 
+struct range {
+  int first_byte;
+  int last_byte;
+  int bytes_covered;
+  int is_assigned;
+  bool *resources;
+  uerr_t status_least_severe;
+};
+
 /* 2005-02-19 SMS.
    Select an appropriate "orig" suffix and a separator character for
    adding a unique suffix to a file name.
